from django.db import models, connection
from django.core.serializers.json import DjangoJSONEncoder


class PublishedEventQueue(models.Model):
    id = models.IntegerField(primary_key=True)
    create_time = models.DateTimeField(auto_now_add=True)

    is_new_style = models.BooleanField(null=True)

    exchange = models.CharField(max_length=64)
    routing_key = models.CharField(max_length=256)
    body = models.JSONField(null=True, encoder=DjangoJSONEncoder)

    event_time = models.DateTimeField(null=True)
    aggregate = models.CharField(null=True, max_length=64)
    event_type = models.CharField(null=True, max_length=64)
    payload = models.JSONField(null=True, encoder=DjangoJSONEncoder)

    send_after_time = models.DateTimeField(null=True)
    tags = models.JSONField(null=True)

    properties = models.JSONField(null=True)

    @classmethod
    def get_next_id_value(cls):
        with connection.cursor() as cursor:
            cursor.execute(f"SELECT nextval('services_communication_publishedeventqueue_id_seq');")
            result = cursor.fetchone()
        return result[0]


# Generated by Django 5.0 on 2025-09-17 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('services_communication', '0005_publishedeventqueue_rename_chanel_20250313_0733'),
    ]

    operations = [
        migrations.AddField(
            model_name='publishedeventqueue',
            name='properties',
            field=models.JSONField(null=True),
        ),
    ]



from django.utils.timezone import now

from services_communication.utils import camelize
from services_communication import publishing_backend
from services_communication.settings import communication_settings



def _add_event_in_queue(aggregate, event_type, event_time, payload, tags):
    exchange = aggregate
    routing_key = event_type

    entity_id = publishing_backend.get_id()

    body = {
        'eventId': entity_id,
        'event_time': event_time,
        'aggregate': aggregate,
        'event_type': event_type,
        'payload': camelize(payload),
    }

    publishing_backend.save_message(
        id=entity_id,
        send_after_time=event_time,
        exchange=exchange,
        routing_key=routing_key,
        body=body,
        tags=tags,
    )


def publish_aggregate_event(aggregate, event_type, payload):
    _add_event_in_queue(
        aggregate, event_type, now(),
        payload,
        None,
    )


def publish_future_aggregate_event(aggregate, event_type, event_time, payload, tags=None):
    assert communication_settings.PUBLISHER_FUTURE_EVENT_ENABLE, 'Publishing of future event not enabled in settings! Set PUBLISHER_FUTURE_EVENT_ENABLE to True'

    _add_event_in_queue(
        aggregate, event_type, event_time,
        payload,
        normalize_tags(tags),
    )


def cancel_future_aggregate_event(aggregate, event_type, tags):
    assert tags, "Tags cant not be empty for canceling future event"

    norm_tags = normalize_tags(tags)

    publishing_backend.delete_by_tags_subset(aggregate, event_type, norm_tags)

def normalize_tags(tags):
    if not tags:
        return None

    return {k: str(v) for k, v in tags.items()}




from djangorestframework_camel_case.settings import api_settings as camel_case_api_settings
from djangorestframework_camel_case import util


def underscoreize(data):
    if not data:
        return data
    return util.underscoreize(data, **camel_case_api_settings.JSON_UNDERSCOREIZE)


def camelize(data):
    if not data:
        return data
    return util.camelize(data, **camel_case_api_settings.JSON_UNDERSCOREIZE)


class DefaultCorrelationIdHelper:

    def get_correlation_id(self):
        return None

    def set_correlation_id(self, value):
        pass


import logging
from datetime import timedelta

from django.utils.timezone import now

from services_communication import publishing_backend
from services_communication.settings import communication_settings


logger = logging.getLogger('services_communication.publisher.process')


def run_publisher():
    _check_publisher_settings()
    publisher_queue_handler = _get_publisher_queue_handler()
    publisher_queue_handler.run()


def is_publisher_work(max_queue_size=0, max_delay=15):
    created_before_time = now() - timedelta(seconds=max_delay)
    queue_size = publishing_backend.get_messages_for_send_to_broker().filter(event_time__lt=created_before_time).count()

    return queue_size <= max_queue_size


def _get_publisher_queue_handler():
    return build_publisher_queue_handler_by_settings()


def _check_publisher_settings():
    assert communication_settings.BROKER_CONNECTION_PARAMETERS, 'Broker connection not set!'

def build_publisher_queue_handler_by_settings():
    return communication_settings.MESSAGE_PUBLISHER_QUEUE_HANDLER_CLASS()



import logging
import select
import json
import time
from urllib.parse import quote_plus

try:
    import psycopg2
except ImportError:
    pass

from django.conf import settings

from services_communication import publishing_backend
from services_communication.settings import communication_settings
from services_communication.broker import build_publisher_by_settings, BlockedPublisher


logger = logging.getLogger('services_communication.publisher.process')


class SyncPoolingPublisherQueueHandler:

    def __init__(self, **kwargs):
        self.mq = build_publisher_by_settings(BlockedPublisher)

    def run(self):
        while True:
            for message in publishing_backend.get_messages_for_send_to_broker():
                if message.is_new_style:
                    self.mq.publish(
                        exchange=message.exchange,
                        routing_key=message.routing_key,
                        body=json.dumps(message.body),
                        correlation_id=message.properties.get('correlation_id') if message.properties else None,
                    )
                else:
                    self.mq.publish(
                        exchange=message.exchange,
                        routing_key=message.routing_key,
                        body=json.dumps({
                            'eventId': message.id,
                            'eventTime': message.event_time.isoformat(),
                            'eventType': message.event_type,
                            'aggregate': message.aggregate,
                            'payload': message.payload,

                        }),
                    )
                publishing_backend.delete(message)
            time.sleep(1)

class ListenPublisherQueueHandler:

    notification_select_timeout = 15

    def __init__(self, **kwargs):
        assert psycopg2, 'ListenPublisher work by NOTIFY/LISTEN and available only for PostgreSQL and required "psycopg2"'
        assert not communication_settings.PUBLISHER_FUTURE_EVENT_ENABLE, 'ListenPublisher can\'t handel with future event yet'

        db_settings = settings.DATABASES[publishing_backend.get_db_alias()]
        if "postgresql" not in db_settings["ENGINE"]:
            raise Exception("ListenPublisherQueueHandler available only if PublishedEventQueue model use PostgreSQL databases.")
        db_dsn = build_settings_to_url(db_settings)

        scheme, tabel = publishing_backend.get_scheme_and_tabel_name()
        self.chanel_name = f"{scheme}_eventqueued"

        self.mq = build_publisher_by_settings(BlockedPublisher)

        self.conn = psycopg2.connect(db_dsn, **{'async': 1})
        self.wait()

    def publish_messages_from_queue(self):
        for message in publishing_backend.get_messages_for_send_to_broker():
            if message.is_new_style:
                self.mq.publish(
                    exchange=message.exchange,
                    routing_key=message.routing_key,
                    body=json.dumps(message.body),
                    correlation_id=message.properties.get('correlation_id') if message.properties else None,
                )
            else:
                self.mq.publish(
                    exchange=message.exchange,
                    routing_key=message.routing_key,
                    body=json.dumps({
                        'eventId': message.id,
                        'eventTime': message.event_time.isoformat(),
                        'eventType': message.event_type,
                        'aggregate': message.aggregate,
                        'payload': message.payload,

                    }),
                )
            publishing_backend.delete(message)

    def run(self):

        logger.info('Chane name for listen pg notification: {}'.format(self.chanel_name))

        with self.conn.cursor() as curs:
            curs.execute(f"LISTEN {self.chanel_name};")
            if not self.wait(socket_timeout=5):
                raise psycopg2.OperationalError('Set listen statement execution soket await timeout')
            logger.debug('Listener set')

        self.publish_messages_from_queue()

        while True:
            timeout_counter = 0
            while timeout_counter < 60:
                has_notifications = self.await_notifies()
                if has_notifications:
                    timeout_counter = 0
                    self.publish_messages_from_queue()
                else:
                    timeout_counter = timeout_counter + self.notification_select_timeout

            logger.debug('Start connection health status')
            if not self.is_connection_healthy():
                logger.error('Response do not received in timeout. Connection unhealthy')
                self.conn.close()
                raise psycopg2.OperationalError('Connection unhealthy!')
            logger.debug('Connection healthy')

    def await_notifies(self):
        logger.debug("Await notifies from DB...")
        if select.select([self.conn], [], [], self.notification_select_timeout) == ([], [], []):
            return False
        else:
            self.conn.poll()
            logger.debug("Receive %d notifications from DB", len(self.conn.notifies))
            self.conn.notifies.clear()
            return True

    def is_connection_healthy(self, timeout=None):
        if not timeout:
            timeout = 5

        with self.conn.cursor() as curs:
            curs.execute("select 1;")
            if not self.wait(socket_timeout=timeout):
                return False
            return True

    def wait(self, socket_timeout=None):
        timeout_pass = False
        last_state = None
        while True:
            state = self.conn.poll()

            if last_state != state:
                timeout_pass = False

            if timeout_pass:
                return False

            if state == psycopg2.extensions.POLL_OK:
                break
            elif state == psycopg2.extensions.POLL_WRITE:
                if select.select([], [self.conn.fileno()], [], socket_timeout) == ([], [], []):
                    timeout_pass = True
            elif state == psycopg2.extensions.POLL_READ:
                if select.select([self.conn.fileno()], [], [], socket_timeout) == ([], [], []):
                    timeout_pass = True
            else:
                raise psycopg2.OperationalError("poll() returned %s" % state)
            last_state = state

        return True


def build_settings_to_url(db_settings):
    DATABASE_URL = f"postgresql://{quote_plus(db_settings['USER'])}:{quote_plus(db_settings['PASSWORD'])}@{db_settings['HOST']}:{db_settings['PORT']}/{db_settings['NAME']}"
    return DATABASE_URL


class SilentRaiseWrapper:

    def __init__(self, function, on_error_callback=None, on_success_callback=None):
        self._function = function
        self._on_error_callback = on_error_callback
        self._on_success_callback = on_success_callback

    def on_error(self, func):
        self._on_error_callback = func
        return self

    def on_success(self, func):
        self._on_success_callback = func
        return self

    def call(self, *args, **kwargs):
        return self(*args, **kwargs)

    def _do_on_error(self, e):
        if self._on_error_callback:
            self._on_error_callback(e)

    def _do_on_success(self, data):
        if self._on_success_callback:
            self._on_success_callback(data)

    def __call__(self, *args, **kwargs):
        try:
            res = self._function(*args, **kwargs)
        except Exception as e:
            self._do_on_error(e)
            return None, e

        self._do_on_success(res)
        return res, None



from django.db import connections
from django.db.models import Q
from django.utils.timezone import now

from services_communication.models import PublishedEventQueue
from services_communication.settings import communication_settings


def get_messages_for_send_to_broker():
    qs = PublishedEventQueue.objects.filter(Q(send_after_time__lte=now()) | Q(event_time__lte=now()))
    return qs

def get_id():
    return PublishedEventQueue.get_next_id_value()

def save_message(id, exchange, routing_key, send_after_time, body, tags):
    PublishedEventQueue.objects.create(
        id=id,
        send_after_time=send_after_time,
        exchange=exchange,
        routing_key=routing_key,
        body=body,
        is_new_style=True,
        tags=tags,
        properties=dict(
            correlation_id=communication_settings.CORRELATION_ID_HELPER.get_correlation_id(),
        )
    )


def delete(message):
    message.delete()


def delete_by_tags_subset(exchange, routing_key, tags):
    PublishedEventQueue.objects.filter(exchange=exchange, routing_key=routing_key, tags__contains=tags).delete()


def get_db_alias():
    return PublishedEventQueue.objects.db


def get_scheme_and_tabel_name():
    table = PublishedEventQueue._meta.db_table

    connection = connections[get_db_alias()]
    with connection.cursor() as cursor:
        cursor.execute(
            """SHOW search_path;""",
        )

        search_path = cursor.fetchone()[0]
        if not search_path:
            raise Exception('Database search_path is empty!')

        search_path_schemes = search_path.split(',')

    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT table_schema FROM information_schema.tables WHERE table_name = %s AND table_schema = ANY(%s);""",
            (table, search_path_schemes),
        )

        schemes_has_tabel = set(row[0] for row in cursor.fetchall())

    schema = None
    for available_scheme in search_path_schemes:
        if available_scheme in schemes_has_tabel:
            schema = available_scheme
            break

    if not schema:
        raise Exception('Not found db schema used for tabel {}!'.format(table))

    return schema, table


# Generated by Django 4.2.17 on 2025-03-07 11:43

from django.db import migrations, router


def create_trigger_function(apps, schema_editor):
    PublishedEventQueue = apps.get_model("services_communication", "PublishedEventQueue")
    if not router.allow_migrate_model(schema_editor.connection.alias, PublishedEventQueue):
        return

    if schema_editor.connection.vendor != 'postgresql':
        return

    schema_editor.execute("""
        CREATE OR REPLACE FUNCTION notify_insert_row() RETURNS TRIGGER AS $$
        BEGIN
            PERFORM pg_notify('new_' || TG_TABLE_SCHEMA::text || '_' || TG_TABLE_NAME::text, null);
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
    """)

    schema_editor.execute(f"""
        CREATE TRIGGER notify_new_published_event_trigger
        AFTER INSERT ON {PublishedEventQueue._meta.db_table}
        FOR EACH ROW EXECUTE FUNCTION notify_insert_row();
    """)



class Migration(migrations.Migration):

    dependencies = [
        ('services_communication', '0003_publishedeventqueue_body_and_more'),
    ]

    operations = [
        migrations.RunPython(create_trigger_function),
    ]



import pytz
from datetime import datetime
from functools import wraps

from django.db import connections
from django.db.transaction import atomic
from django.utils.timezone import now

from services_communication.utils import camelize
from services_communication import publishing_backend
from services_communication.defaults import message_router as _message_router
message_router = _message_router


def event_consumer(f):
    @wraps(f)
    def wrapper(routing_key, message_body, *args, **kwargs):
        event_id = message_body.get('event_id')
        event_time = message_body.get('event_time')
        event_aggregate = message_body.get('aggregate')
        event_event_type = message_body.get('event_type')
        payload = message_body.get('payload')
        return f(payload,
                 event_id=event_id,
                 event_time=event_time,
                 aggregate=event_aggregate,
                 event_type=event_event_type)

    return wrapper


def command_consumer(f):
    @wraps(f)
    def wrapper(routing_key, message_body, *args, **kwargs):
        meta = message_body.pop('meta', {})

        expired_time_str = meta.get('expired_time')
        if expired_time_str:
            expired_time = pytz.utc.localize(datetime.strptime(expired_time_str, '%Y-%m-%dT%H:%M:%S.%fZ'))
            if expired_time < now():
                return

        return f(routing_key, message_body, *args, **kwargs)

    return wrapper



def call_consumer(func=None, *, no_atomic=False, check_expired_time_after_handel=None):
    def decorator(f):
        @wraps(f)
        def wrapper(routing_key, message_body, *args, exchange=None, **kwargs):
            call_id = message_body['id']
            request_meta = message_body.pop('meta', {})
            request_data = message_body.pop('data', {})


            expired_time = None
            expired_time_str = request_meta.get('expired_time')
            if expired_time_str:
                expired_time = pytz.utc.localize(datetime.strptime(expired_time_str, '%Y-%m-%dT%H:%M:%S.%fZ'))
                if expired_time < now():
                    return


            response = f(request_data, *args, method_name=routing_key, call_id=call_id, meta=request_meta, **kwargs)
            # Если тут поймать ошибку и не прокинуть ее выше, то не произойдет откат транзакции, а это не допустимо.
            # Кроме того, как в это случае "возвращать" ответ, ведь ответ то тоже требует записи в базу.

            if check_expired_time_after_handel and expired_time:
                if expired_time < now():
                    for db in connections.all():
                        db.set_rollback(True)
                    return

            reply_to = request_meta['reply_to']
            response_meta = {
                'time': now(),
            }

            body = {
                'id': call_id,
                'data': response,
                'meta': response_meta,
            }

            publishing_backend.save_message(
                id=publishing_backend.get_id(),
                send_after_time=now(),
                exchange=reply_to,
                routing_key=f'call-response.{exchange}.{routing_key}',
                body=camelize(body),
                tags=None,
            )
            return

        if no_atomic:
            return wrapper
        return atomic(wrapper)

    if func and callable(func):
        return decorator(func)
    return decorator


def call_response_consumer(f):
    @wraps(f)
    def wrapper(routing_key, message_body, *args, exchange=None, **kwargs):
        call_id = message_body['id']
        request_meta = message_body.pop('meta', {})
        request_data = message_body.pop('data', {})

        _, service_name, method_name = routing_key.split('.', 2)

        response = f(request_data, *args, service_name=service_name, method_name=method_name, call_id=call_id, meta=request_meta, **kwargs)
    return wrapper



# Generated by Django 4.2.2 on 2023-06-14 16:39

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PublishedEventQueue',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('exchange', models.CharField(max_length=64)),
                ('routing_key', models.CharField(max_length=256)),
                ('event_time', models.DateTimeField(auto_now_add=True)),
                ('aggregate', models.CharField(max_length=64)),
                ('event_type', models.CharField(max_length=64)),
                ('payload', models.JSONField()),
            ],
        ),
    ]



# Generated by Django 4.2.17 on 2024-12-18 10:31

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('services_communication', '0002_publishedeventqueue_create_time_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='publishedeventqueue',
            name='body',
            field=models.JSONField(encoder=django.core.serializers.json.DjangoJSONEncoder, null=True),
        ),
        migrations.AddField(
            model_name='publishedeventqueue',
            name='is_new_style',
            field=models.BooleanField(null=True),
        ),
        migrations.AddField(
            model_name='publishedeventqueue',
            name='send_after_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='aggregate',
            field=models.CharField(max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='event_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='event_type',
            field=models.CharField(max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='payload',
            field=models.JSONField(encoder=django.core.serializers.json.DjangoJSONEncoder, null=True),
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name='publishedeventqueue',
                    name='id',
                    field=models.IntegerField(primary_key=True, serialize=False),
                ),
            ],
        )
    ]



import logging
import random
import time

import pika
from pika.exceptions import AMQPConnectionError
from pika.exchange_type import ExchangeType

from services_communication.error import MessageNotConsumed
from services_communication.settings import communication_settings
from services_communication.types import Exchange

logger_consumer = logging.getLogger('services_communication.consumer')
logger_publisher = logging.getLogger('services_communication.publisher')


class BlockedMixin:

    def _declare_queue(self, channel):
        if communication_settings.DO_NOT_DECLARE:
            return
        channel.queue_declare(self._queue, durable=True)

    def _declare_exchanges(self, channel):
        if communication_settings.DO_NOT_DECLARE:
            return
        for exchange in self._exchanges:
            channel.exchange_declare(exchange.name, exchange_type=exchange.type, durable=True)

    def _declare_binds(self, channel):
        if communication_settings.DO_NOT_DECLARE:
            return
        for bind in self._binds:
            channel.queue_bind(self._queue, bind.exchange, routing_key=bind.routing_key)


class BlockedConsumer(BlockedMixin):

    def __init__(self,
                 broker_connection_parameters=None,
                 queue=None,
                 exchanges=None,
                 binds=None,
                 on_message_callback=None,
                 ignore_callback_error=False,
                 **kwargs):
        logger_consumer.info('Init {} with callback {}'.format(type(self).__name__, on_message_callback))
        self._broker_connection_parameters = broker_connection_parameters
        self._queue = queue
        self._exchanges = exchanges
        self._binds = binds
        self._on_message_callback = on_message_callback
        self._ignore_callback_error = ignore_callback_error

    def on_message(self, channel, method_frame, header_frame, body):
        logger_consumer.debug('Received message (dt:{}) from exchange {} with key {}'.format(method_frame.delivery_tag,
                                                                                             method_frame.exchange,
                                                                                             method_frame.routing_key))
        try:
            communication_settings.CORRELATION_ID_HELPER.set_correlation_id(header_frame.correlation_id)
            self._on_message_callback(method_frame, header_frame, body)
        except MessageNotConsumed as e:
            logger_consumer.warning('Message (dt:{}) from {}:{} not consume'.format(method_frame.delivery_tag, method_frame.exchange, method_frame.routing_key))
            channel.basic_reject(delivery_tag=method_frame.delivery_tag, requeue=True)
            return
        except Exception as e:
            logger_consumer.exception('Message (dt:{}) from {}:{} consume raise error'.format(method_frame.delivery_tag, method_frame.exchange, method_frame.routing_key))
            channel.basic_reject(delivery_tag=method_frame.delivery_tag, requeue=True)
            if self._ignore_callback_error:
                return
            raise
        channel.basic_ack(delivery_tag=method_frame.delivery_tag)

    def declair(self, channel):
        self._declare_queue(channel)
        self._declare_exchanges(channel)
        self._declare_binds(channel)

    def run(self):
        try:
                logger_consumer.info("Connect to rabbit")
                with pika.BlockingConnection(parameters=self._broker_connection_parameters) as connection:
                    logger_consumer.info("Open channel")
                    with connection.channel() as channel:
                        # channel.basic_qos(prefetch_count=15)
                        self.declair(channel)

                        channel.basic_consume(self._queue, self.on_message)
                        try:
                            logger_consumer.info("Starting consuming")
                            channel.start_consuming()
                        except KeyboardInterrupt:
                            logger_consumer.info("Consuming interrupted")
                            channel.stop_consuming()
        except Exception:
            logger_consumer.exception("Error happened in consumer")


class BlockedReconnectingConsumer(BlockedConsumer):

    def __init__(self,
                 **kwargs):
        super().__init__(**kwargs)
        self.max_reconnect_attempts = 3
        self.max_total_reconnect_attempts = self.max_reconnect_attempts * 5
        self.total_reconnect_attempts = 0
        self.is_declared = False

    def on_message(self, *args, **kwargs):
        self.total_reconnect_attempts = 0
        super().on_message(*args, **kwargs)

    def declair(self, channel):
        if not self.is_declared:
            super().declair(channel)
            self.is_declared = True

    def reconnect_pause_random_gen(self):
        yield random.randint(3, 8)

    def run(self):
        self.total_reconnect_attempts = 0
        reconnect_attempts = self.max_reconnect_attempts

        while True:
            try:
                    logger_consumer.info("Connect to rabbit")
                    with pika.BlockingConnection(parameters=self._broker_connection_parameters) as connection:
                        logger_consumer.info("Open channel")
                        with connection.channel() as channel:
                            reconnect_attempts = 0
                            # channel.basic_qos(prefetch_count=15)
                            self.declair(channel)

                            channel.basic_consume(self._queue, self.on_message)
                            try:
                                logger_consumer.info("Starting consuming")
                                channel.start_consuming()
                            except KeyboardInterrupt:
                                logger_consumer.info("Consuming interrupted")
                                channel.stop_consuming()
                                break
            except AMQPConnectionError as err:
                self.total_reconnect_attempts = self.total_reconnect_attempts + 1
                reconnect_attempts = reconnect_attempts + 1
                allow_reconnect = reconnect_attempts <= self.max_reconnect_attempts and self.total_reconnect_attempts <= self.max_total_reconnect_attempts
                if allow_reconnect:
                    logger_consumer.warning("Connection error on %s attempt: %s", reconnect_attempts, err)
                    pause = next(self.reconnect_pause_random_gen())
                    logger_consumer.warning('Await {} before reconnect attempt'.format(pause))
                    time.sleep(pause)
                else:
                    logger_consumer.exception("Connection error on %s attempt: %s. Can not connect to broker", reconnect_attempts, err)
                    break
            except Exception:
                logger_consumer.exception("Error happened in consumer")
                break


class BlockedPublisher(BlockedMixin):

    def __init__(self,
                 app_id=None,
                 broker_connection_parameters=None,
                 exchanges=None,
                 **kwargs):
        self._broker_connection_parameters = broker_connection_parameters
        self._exchanges = exchanges
        self._connection = None
        self._channel = None
        self._app_id = app_id or 'anonymous'

    def _publish(self,
                 exchange,
                 routing_key,
                 body,
                 correlation_id=None,
                 **kwargs):

        if not self._connection:
            logger_publisher.info("Connect to rabbit")
            self._connection = pika.BlockingConnection(parameters=self._broker_connection_parameters)
        if not self._channel:
            logger_publisher.info("Open channel")
            self._channel = self._connection.channel()
            self._channel.confirm_delivery()

        self._declare_exchanges(self._channel)

        return self._channel.basic_publish(
            exchange,
            routing_key,
            body,
            properties=pika.BasicProperties(
                content_type='text/json',
                delivery_mode=pika.DeliveryMode.Persistent,
                app_id=self._app_id,
                correlation_id=correlation_id,
            ),
        )

    def publish(self, *args, **kwargs):
        try:
            self._publish(*args, **kwargs)
        except pika.exceptions.ConnectionClosedByBroker as err:
            # Uncomment this to make the example not attempt recovery
            # from server-initiated connection closure, including
            # when the node is stopped cleanly
            #
            # break
            logger_publisher.warning("Caught a connection error: {}, retrying...".format(err))
            self._retry_publish(*args, **kwargs)
        # Do not recover on channel errors
        except pika.exceptions.AMQPChannelError as err:
            logger_publisher.exception("Caught a channel error: {}, stopping...".format(err))
            raise err
        # Recover on all other connection errors
        except pika.exceptions.AMQPConnectionError:
            logger_publisher.warning("Connection was closed, retrying...")
            self._retry_publish(*args, **kwargs)
        except Exception as err:
            logger_publisher.exception("Caught a error: {}, stopping...".format(err))
            raise err

    def _retry_publish(self, *args, **kwargs):
        logger_publisher.warning("Retrying publish...")
        self._channel = None
        self._connection = None
        self._publish(*args, **kwargs)


class OneMessagePublisher(BlockedPublisher):

    def publish(self, exchange, *args, **kwargs):
        self._exchanges = [Exchange(exchange, ExchangeType.topic)]

        self._publish(exchange, *args, **kwargs)

        if self._channel and self._channel.is_open:
            self._channel.close()
        if self._connection and self._connection.is_open:
            self._connection.close()



def build_publisher_by_settings(publisher_class):
    return publisher_class(
        app_id=communication_settings.APP_ID,
        broker_connection_parameters=communication_settings.BROKER_CONNECTION_PARAMETERS,
        exchanges=communication_settings.EXCHANGES,
    )



import json
from datetime import timedelta

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import now

from services_communication.settings import communication_settings
from services_communication.broker import OneMessagePublisher, build_publisher_by_settings
from services_communication.utils import camelize


def send_command(service_name, arguments, command_name='', timeout=None, expired_time=None):
    assert not (timeout and expired_time), "Ambiguous timeout! Use timeout or expired_time, not both"
    assert expired_time is None or expired_time > now(), "expired_time in past!"

    meta = {}

    if timeout:
        expired_time = now() + timedelta(seconds=timeout)

    if expired_time:
        meta['expired_time'] = expired_time

    arguments['meta'] = meta

    build_publisher_by_settings(OneMessagePublisher).publish(
        exchange=service_name,
        routing_key=command_name,
        body=json.dumps(camelize(arguments), cls=DjangoJSONEncoder),
        correlation_id=communication_settings.CORRELATION_ID_HELPER.get_correlation_id(),
    )



import requests
from django.utils.translation import get_language

from services_communication.rest_api import error
from services_communication.settings import communication_settings

from services_communication.rest_api.formatter import json_request, json_response, full_response
from services_communication.rest_api.auth_helper import get_alive_access_token

API_HOST = communication_settings.REST_API_HOST


def get(uri, request_formatter=json_request, response_formatter=json_response, **kwargs):
    return _request(uri, 'GET', request_formatter, response_formatter, **kwargs)


def post(uri, request_formatter=json_request, response_formatter=json_response, **kwargs):
    return _request(uri, 'POST', request_formatter, response_formatter, **kwargs)


def put(uri, request_formatter=json_request, response_formatter=json_response, **kwargs):
    return _request(uri, 'PUT', request_formatter, response_formatter, **kwargs)


def patch(uri, request_formatter=json_request, response_formatter=json_response, **kwargs):
    return _request(uri, 'PATCH', request_formatter, response_formatter, **kwargs)


def delete(uri, request_formatter=json_request, response_formatter=full_response, **kwargs):
    return _request(uri, 'DELETE', request_formatter, response_formatter, **kwargs)


def head(uri, request_formatter=json_request, **kwargs):
    return _request(uri, 'HEAD', request_formatter, full_response, **kwargs)


def _request(uri, method, request_formatter, response_formatter, params={}, json=None, data=None, files=None, headers=None, no_auth=False, extra_host=None, **kwargs):
    url = build_url(uri, extra_host=extra_host)

    if headers is None:
        headers = {}

    headers['Accept-Language'] = get_language()

    if not no_auth:
        headers['Authorization'] = 'Bearer ' + get_alive_access_token()

    try:
        response = requests.request(
            method,
            url,
            headers=headers,

            params=request_formatter(params),
            json=request_formatter(json),
            data=request_formatter(data),
            files=files,
            **kwargs,
        )
    except requests.exceptions.ConnectionError as e:
        raise error.RestApiConnectionError(url, method, e)
    except requests.exceptions.Timeout as e:
        raise error.RestApiTimeoutError(url, method, e)
    except requests.exceptions.RequestException as e:
        raise error.RestApiRequestError(url, method, e)

    if response.status_code // 100 != 2:
        raise build_response_error(uri, method, response)

    return response_formatter(response)


def build_url(uri, extra_host=None):
    host = extra_host if extra_host else API_HOST
    return f'{host}/{uri}'


def build_response_error(url, method, response):
    # todo: add service api error handel
    error_class = ERROR_BY_STATUS_MAP.get(response.status_code)

    if not error_class:
        if response.status_code // 100 == 4:
            error_class = error.RestApiClientError
        elif response.status_code // 100 == 5:
            error_class = error.RestApiServerError
        else:
            error_class = error.RestApiResponseWithError

    return error_class(url, method, response)


ERROR_BY_STATUS_MAP = {
    400: error.RestApiBadRequestError,
    401: error.RestApiUnauthorizedError,
    403: error.RestApiForbiddenError,
    404: error.RestApiNotFountError,
    502: error.RestApiBadGatewayError,
    504: error.RestApiGatewayTimeoutError,
}



from django.core.management import BaseCommand, CommandError
from services_communication.process import is_publisher_work

import logging

logging.basicConfig(level=logging.WARNING)


class Command(BaseCommand):
    help = "Run publisher for queued aggregate event"

    def add_arguments(self, parser):
        parser.add_argument(
            "--max_queue_size",
            type=int,
            default=0,
            help="The maximum number of unhandled events at which a publisher is considered healthy",
        )
        parser.add_argument(
            "--max_delay",
            type=int,
            default=10,
            help="The number of seconds must pass after the event to begin to consider it unprocessed",
        )

    def handle(self, *args, **options):
        is_healthy = is_publisher_work(
            max_queue_size=options['max_queue_size'],
            max_delay=options['max_delay'],
        )

        if not is_healthy:
            raise CommandError('Publisher unhealthy')

        self.stdout.write(self.style.SUCCESS('Publisher healthy'))



from datetime import timedelta

from django.utils.timezone import now

from services_communication.utils import camelize
from services_communication import publishing_backend



def send_call(service_name, method_name, data, reply_to=None, timeout=None, expired_time=None):

    assert reply_to, "Exchange for reply not set!"
    assert not (timeout and expired_time), "Ambiguous timeout! Use timeout or expired_time, not both"
    assert expired_time is None or expired_time > now(), "expired_time in past!"

    meta = {
        'reply_to': reply_to,
        'time': now(),
    }

    if timeout:
        expired_time = now() + timedelta(seconds=timeout)

    if expired_time:
        meta['expired_time'] = expired_time

    call_id = publishing_backend.get_id()

    body = {
        'id': call_id,
        'meta': meta,
        'data': data,
    }

    publishing_backend.save_message(
        id=call_id,
        send_after_time=now(),
        exchange=service_name,
        routing_key=method_name,
        body=camelize(body),
        tags=None,
    )

    return call_id


# Generated by Django 4.2.2 on 2024-03-13 12:32

import django.core.serializers.json
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('services_communication', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='publishedeventqueue',
            name='create_time',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='publishedeventqueue',
            name='tags',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='event_time',
            field=models.DateTimeField(),
        ),
        migrations.AlterField(
            model_name='publishedeventqueue',
            name='payload',
            field=models.JSONField(encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
    ]
